unit mIRCStuff;

interface

uses
  Windows, Messages;

const
  WhiteSpaceChars = [#9, #10, #13, #32];
  MIRC_STRING = 900;  

type
  // Structure for LoadDll()/UnloadDll() routines.
  LoadInfo = packed record
    mVersion: DWORD;
    mHwnd: HWND;
    mKeep: BOOL;
  end;

  PLoadInfo =  ^LoadInfo;

  LongRec = packed record
    case Integer of
      0: (Lo, Hi: Word);
    end;

  TmString = array[0..MIRC_STRING] of Char;

const
  // Return codes for functions.
  S_OK = 'S_OK';
  E_INTEGER = 'E_INVALID_INTEGER';

  // Message for evaluating mIRC commands.
  WM_MCOMMAND = WM_USER + 200;

  // Messages used internally so that tbwin.dll and tbtray.dll work together.
  WM_NOTIFYTBTRAY = WM_USER + 1000;
  WM_NOTIFYTBWIN  = WM_USER + 1001;

var
  MainHandle: HWND;

function trim(S: string): string;
function gettok(Source: string; const C: Char; N: Integer): string;
function intstr(const Value: Integer): string;
function upper(const Value: string): string;
function mstr(const Value: string): TmString;
function makelong(const Hi, Lo: Word): Longint;

function NewMessageProc(const Handle: HWND; Proc: Pointer): Pointer;
function RestoreMessageProc(const Handle: HWND; Proc: Pointer): Pointer;

procedure Return(data: PChar; const S: string);

procedure ExecCommand(const S: string);

implementation

// Copies S into data.
procedure Return(data: PChar; const S: string);
begin
  lstrcpy(data, PChar(S));
end;


function makelong(const Hi, Lo: Word): Longint;
begin
  LongRec(Result).Hi := Hi;
  LongRec(Result).Lo := Lo;
end;

function mstr(const Value: string): TmString;
var
  I, Len: Integer;
begin
  Len := Length(Value);
  if (Len > MIRC_STRING) then
    Len := MIRC_STRING;

  I := 0;
  
  while (I < Len) do
  begin
    Result[I] := Value[I + 1];
    Inc(I);
  end;

  Result[I] := #0;
end;

// Sets the window procedure for Handle to Proc and returns a pointer
// to the old one.  Returns nil if the handle doesn't exist.
function NewMessageProc(const Handle: HWND; Proc: Pointer): Pointer;
begin
  // Go ahead and set this to nil in case the handle is invalid.
  Result := nil;

  // Make sure the handle is a valid window.  If it is, save the
  // original window procedure and set the new one.
  if (IsWindow(Handle)) then
  begin
    Result := Pointer(GetWindowLong(Handle, GWL_WNDPROC));
    SetWindowLong(Handle, GWL_WNDPROC, Longint(Proc));
  end;

end;

// Restores the message procedure for Handle, and sets the result to
// nil so you can set the original pointer to nil.
function RestoreMessageProc(const Handle: HWND; Proc: Pointer): Pointer;
begin
  // Set the result to nil.
  Result := nil;

  // If the window is valid, restore the original window procedure.
  // This is usually done when you're unloading the dll.
  if (IsWindow(Handle)) then
    SetWindowLong(Handle, GWL_WNDPROC, Longint(Proc));
end;

procedure ExecCommand(const S: string);
var
  MapFile: Integer;
  Command: Pointer;
begin
  // Create a memory mapped file.
  MapFile := CreateFileMapping(INVALID_HANDLE_VALUE, nil, PAGE_READWRITE, 0, 1024, 'mIRC');

  // Basically, allocate (SIZE OF FILE) memory into the pointer
  // so that we can copy a command into it.  1024 is good enough
  // because mIRC only supports strings that are ~900 chars long.
  Command := MapViewOfFile(MapFile, FILE_MAP_ALL_ACCESS, 0, 0, 0);

  // Copy the command into the pointer to the memory mapped file.
  lstrcpy(Command, PChar(S));

  // Tell mIRC "yo, we got something for you to do." :)
  SendMessage(MainHandle, WM_MCOMMAND, 1, 0);

  // Release the memory used by the pointer for the command.
  UnmapViewOfFile(Command);

  // Close the memory mapped file.
  CloseHandle(MapFile);
end;

// Returns the specified integer as a string.
function intstr(const Value: Integer): string;
begin
  Str(Value, Result);
end;

// Returns the string in all uppercase.
function upper(const Value: string): string;
var
  I: Integer;
begin
  Result := Value;
  for I := 1 to Length(Result) do
    Result[I] := UpCase(Result[I]);
end;

// Removes excess white space from the beginning and end
// of strings.
function trim(S: string): string;
begin
  Result := S;

  // Remove from front.
  while (Length(S) > 0) and (S[1] in WhiteSpaceChars) do
    Delete(S, 1, 1);

  // Remove from back.
  while (Length(S) > 0) and (S[Length(S)] in WhiteSpaceChars) do
    Delete(S, Length(S), 1);
end;

// gettok(source, token, N)
// if N is negative, it returns from N onwards (equiv. of N- in mIRC).
function gettok(Source: string; const C: Char; N: Integer): string;
var
  SPos, Num: Integer;
  All: Boolean;
begin
  Result := '';

  All := (N < 0);

  if (N < 0) then
    N := N * -1;

  if (Source = '') then
    Exit;

  while (Source[1] = C) do
    Source := Copy(Source, 2, MaxInt);
  while (Source[Length(Source)] = C) do
    Source := Copy(Source, 1, Length(Source) - 1);

  Num := 1;
  SPos := Pos(C, Source);
  while (SPos > 0) do
  begin
    if (Num = N) then
    begin
      if (All) then
        Result := trim(Copy(Source, 1, MaxInt))
      else
        Result := trim(Copy(Source, 1, SPos - 1));
      Exit;
    end;

    Delete(Source, 1, SPos);
    Inc(Num);
    SPos := Pos(C, Source);
  end;

  if (Num = N) then
    Result := trim(Source);
end;

end.
