library tbtray;

uses
  Windows, Messages, mIRCStuff;

const
  BORDER_SOLID = 65535;

  TT_TIMER = 7777;

type
  TFontStyle = (fsBold, fsItalic, fsUnderline, fsStrikeout);
  TFontStyles = set of TFontStyle;

  // Structure to hold information about the tray.
  TTrayInfo = record
    Color: Integer;
    BkColor: Integer;
    FontName: string;
    FontSize: Integer;
    FontStyle: TFontStyles;
    Text: string;
    Transparent: Boolean;
    Width: Integer;
    MarginTop,
    MarginRight,
    MarginBottom: Integer;
    BorderColor: Integer;
    BorderStyle: UINT;
    UpdateRect: TRect;
  end;

const
  TB_MenuStart = 35000;

var
  TB_Handle: HWND;
  TB_WndProc: Pointer = nil;
  TB_OldProc: Pointer = nil;
  TB_ClassStyle: Integer;
  TB_TrayInfo: TTrayInfo;
  TB_Menu: HMENU;
  TB_MenuItems: array[1..256] of string;
  TB_MenuItemCount: Integer;
  TB_MenuDclick: string;
  TB_MenuSclick: string;

  // Tooltip junk...
  TT_Handle: HWND;
  TT_Text: string;
  TT_Delay: Integer;
  TT_Size: TSize;

  LoadedFirst: Boolean = False;
  NoMouseMove: Boolean = False;
  NeedUnload: Boolean;

// Forward declarations
function TB_SubClass(hWnd: HWND; Msg: Cardinal; wParam: Integer; lParam: Integer): Integer; stdcall; forward;
procedure TB_ClearMenu; forward;
procedure TB_Update; forward;
procedure TB_UpdatePaint; forward;
function TB_GetUpdateRect: TRect; forward;
function TB_Initialize(mWnd: HWND): string; forward;
function TB_Uninitialize: string; forward;

procedure TT_Show; forward;
procedure TT_Hide; forward;
procedure TT_Update(P: TPoint); forward;

procedure LoadDll(Info: PLoadInfo); stdcall;
begin

{******* mIRC bugfix! *******}
  if (IsWindow(MainHandle)) and (MainHandle = Info.mHwnd) then
  begin
    Info.mKeep := False;
    NeedUnload := True;
    Exit;
  end;
{****************************}

  // Save the main mIRC handle.  We'll need it.
  MainHandle := Info.mHwnd;

  // Initialize these to zero values.
  TB_Handle   := 0;
  TB_OldProc  := nil;
  SetRect(TB_TrayInfo.UpdateRect, 0, 0, 0, 0);

  // Create the menu and initialize everything.
  TB_Menu := CreatePopupMenu;
  TB_MenuItemCount := 0;
  TB_MenuDclick := '';
  TB_MenuSclick := '';

  // Set the default values for the tooltip (disabled).
  TT_Handle := 0;
  TT_Delay := 200;
  TT_Text := '';

  // Intialize our info structure to some
  // nice default values.
  with TB_TrayInfo do
  begin
    Text := 'tbtray';
    FontName := 'Tahoma';
    FontSize := -10;
    FontStyle := [];
    Color := GetSysColor(COLOR_BTNTEXT);
    BkColor := GetSysColor(COLOR_BTNFACE);
    Transparent := False;
    Width := 75;
    MarginTop := 5;
    MarginBottom := 4;
    MarginRight := 4;
    BorderColor := 0;
    BorderStyle := BDR_SUNKENOUTER;
  end;
end;

function UnloadDll(mTimeout: Integer): Integer; stdcall;
begin
  Result := 0;

{******* mIRC bugfix! *******}
  if (NeedUnload) then
  begin
    NeedUnload := False;
    Exit;
  end;
{****************************}

  if (mTimeout = 0) then
    TB_Uninitialize;
end;

function OpenTray(mWnd, aWnd: HWND; data, parms: PChar; show, nopause: BOOL): Integer; stdcall;
var
  Init: string;
begin
  Result := 3;

  // Initialize the tray! :))
  Init := TB_Initialize(mWnd);
  Return(data, Init);
end;

function CloseTray(mWnd, aWnd: HWND; data, parms: PChar; show, nopause: BOOL): Integer; stdcall;
var
  Uninit: string;
begin
  Result := 3;

  // Close the tray. :((
  Uninit := TB_Uninitialize;
  Return(data, Uninit);
end;

function TT_Proc(hWnd: HWND; Msg: Cardinal; wParam: Integer; lParam: Integer): Integer; stdcall;
var
  R, CR: TRect;
  DC: HDC;
  NC_METRICS: NONCLIENTMETRICS;
  X: HFONT;
  P: TPoint;
begin
  // Act like we're gonna handle it!
  Result := 0;

  // This checks to see if the mouse is still within the rectangle and that the
  // active window is still mIRC.  If either test fails, the tooltip is hidden.
  if (Msg = WM_TIMER) then
  begin
    GetCursorPos(P);
    ScreenToClient(TB_Handle, P);

    if not (PtInRect(TB_GetUpdateRect, P)) or (GetForegroundWindow <> MainHandle) then
      TT_Hide;

    Exit;
  end;

  // Let windows go ahead and do the default processing, which won't be much. :)
  Result := DefWindowProc(hWnd, Msg, wParam, lParam);

  if (Msg = WM_PAINT) then
  begin

    // Most of the stuff here works the same as the painting routine for the
    // tray itself.  The exceptions are the fonts and the colors.  Since we want
    // to look like a regular tooltip (same colors), we get those values from
    // the system.  Windows uses the same font for tooltips as it does for
    // status bars.
    NC_METRICS.cbSize := SizeOf(NC_METRICS);
    SystemParametersInfo(SPI_GETNONCLIENTMETRICS, 0, @NC_METRICS, 0);
    X := CreateFontIndirect(NC_METRICS.lfStatusFont);

    GetClientRect(TT_Handle, R);

    DC := GetDC(TT_Handle);

    if (IsWindowVisible(TT_Handle)) then
    begin
      // Get the appropriate colors for the tooltip text and background.
      Windows.SetBkColor(DC, GetSysColor(COLOR_INFOBK));
      Windows.SetTextColor(DC, GetSysColor(COLOR_INFOTEXT));
    end;

    X := SelectObject(DC, X);

    // DrawText with DT_CALCRECT can figure out how big our rectangle needs to
    // be when given some text.  When using DT_CALCRECT, no text is actually
    // drawn, so we need to call it again later to draw the text.
    SetRect(CR, 0, 0, 0, 0);
    DrawText(DC, PChar(TT_Text), Length(TT_Text), CR, DT_NOPREFIX or DT_CALCRECT);
    TT_Size.cx := CR.Right;
    TT_Size.cy := CR.Bottom;

    if (IsWindowVisible(TT_Handle)) then
    begin
      // Fill the background of the tooltip with the right color.
      FillRect(DC, R, HBRUSH(COLOR_INFOBK + 1));

      OffsetRect(CR, 4, 3);

      // Draw the text.
      DrawText(DC, PChar(TT_Text), Length(TT_Text), CR, DT_NOPREFIX);

      // This makes the tooltip have a little 3D edge to it.  I think it looks
      // better than just the regular flat tooltip.
      DrawEdge(DC, R, BDR_RAISEDOUTER, BF_RECT);
    end;

    Inc(TT_Size.cx, 8);
    Inc(TT_Size.cy, 6);

    X := SelectObject(DC, X);
    DeleteObject(X);

    ReleaseDC(TT_Handle, DC);
  end;
end;

procedure TT_Show;
begin
  // This function creates our tooltip window.  Instead of using the built-in
  // windows tooltip controls, I am just using the class STATIC, which provides
  // us with a blank "palette" to draw on.
  TT_Handle := CreateWindowEx(
    WS_EX_TOOLWINDOW,
    'STATIC',
    '',
    WS_POPUP,
    0, 0, 0, 0, 0, 0, hInstance, nil
  );

  // Initially the size is 0.
  TT_Size.cx := 0;
  TT_Size.cy := 0;

  // These are class styles of tooltips.. the CS_SAVEBITS is the key one, it
  // allows us to hide the tooltip without the window underneath receiving a
  // WM_PAINT message, which saves resources a little.
  SetClassLong(TT_Handle, GCL_STYLE, CS_SAVEBITS or CS_VREDRAW or CS_HREDRAW);

  // Subclass the tooltip window.
  SetWindowLong(TT_Handle, GWL_WNDPROC, Integer(@TT_Proc));
end;

procedure TT_Hide;
begin
  // Remove the timers pertaining to the tooltip.
  KillTimer(TT_Handle, 1);
  KillTimer(TB_Handle, TT_TIMER);

  // If it exists, destroy the window...
  if (IsWindow(TT_Handle)) then
    DestroyWindow(TT_Handle);

  // Zero the handle.
  TT_Handle := 0;
end;

procedure TT_Update(P: TPoint);
var
  R: TRect;
begin
  // Convert the point to screen coordinates.
  ClientToScreen(TB_Handle, P);

  // Calculate the width/height of the tooltip.  Our specialized WM_PAINT
  // handler does this for us. :)
  SendMessage(TT_Handle, WM_PAINT, 0, 0);

  // We need the width of the desktop..
  GetClientRect(GetDesktopWindow, R);

  // .. to make sure the tooltip doesn't get chopped off if it's too wide.
  if (P.x + TT_Size.cx > R.Right) then
    P.x := R.Right - TT_Size.cx;

  // Show the tooltip about 18 pixels below the point of the mouse cursor.
  SetWindowPos(TT_Handle, HWND_TOPMOST, P.x, P.y + 18, TT_Size.cx, TT_Size.cy, SWP_NOSENDCHANGING or SWP_SHOWWINDOW or SWP_NOACTIVATE);

  // Repaint!
  UpdateWindow(TT_Handle);
end;

// AddMenu <"caption"|dclick|sclick> <command>
function AddMenu(mWnd, aWnd: HWND; data, parms: PChar; show, nopause: BOOL): Integer; stdcall;
var
  S, MenuText, MenuCommand: string;
begin
  Result := 3;
  S := string(data);

  // If the first word is 'sclick' or 'dclick', then we
  // know these are special menu items and we need to
  // set them differently.
  if (Copy(S, 1, 6) = 'dclick') then
    TB_MenuDclick := Copy(S, Pos(' ', S) + 1, MaxInt)
  else if (Copy(S, 1, 6) = 'sclick') then
    TB_MenuSclick := Copy(S, Pos(' ', S) + 1, MaxInt)

  // Otherwise, we are adding a right-click menu item.
  else
  begin
    // Make sure they don't add more than 256 menu items... could
    // 'cause problems...
    if (TB_MenuItemCount > 255) then
    begin
      Return(data, 'E_TOO_MANY_MENUS');
      Exit;
    end;

    // We remove the first quote and scan for
    // the second quote.  Everything after the second
    // quote is the command text.  Everything before
    // is the caption of the menu item.
    Delete(S, 1, 1);
    MenuText := Copy(S, 1, Pos('"', S) - 1);
    Delete(S, 1, Pos('"', S) + 1);
    MenuCommand := S;

    // Increase the number of menu items.
    Inc(TB_MenuItemCount);

    // If the caption was '-', we add a separator..
    if (MenuText = '-') then
      InsertMenu(TB_Menu, $FFFFFFFF, MF_BYPOSITION or MF_SEPARATOR, TB_MenuItemCount + TB_MenuStart, '')
    // Otherwise, we add the regular menu item.
    else
    begin
      TB_MenuItems[TB_MenuItemCount] := MenuCommand;
      InsertMenu(TB_Menu, $FFFFFFFF, MF_BYPOSITION, TB_MenuItemCount + TB_MenuStart, PChar(MenuText));
    end;

  end;

  Return(data, S_OK);
end;

// ClearMenu _
function ClearMenu(mWnd, aWnd: HWND; data, parms: PChar; show, nopause: BOOL): Integer; stdcall;
begin
  Result := 3;

  // Free the memory used by the menu and reset everything.
  TB_ClearMenu;

  Return(data, S_OK);
end;

// SetFont [-ibus] <size> <font>
function SetFont(mWnd, aWnd: HWND; data, parms: PChar; show, nopause: BOOL): Integer; stdcall;
var
  Font, Size, Style: string;
  S: string;
  V, E, N: Integer;
begin
  Result := 3;
  S := trim(string(data));

  Style := '';

  N := 1;

  // Check to see if they are using switches or a negative
  // number for the font size.
  if (S[1] = '-') and not (S[2] in ['0'..'9'])  then
  begin
    Style := gettok(S, ' ', 1);
    Inc(N);
  end;

  Size := gettok(S, ' ', N);
  Font := gettok(S, ' ', (N + 1) * -1);

  V := TB_TrayInfo.FontSize;

  if (Size <> '') then
    Val(Size, V, E);

  if (E <> 0) then
    Return(data, 'E_INVALID_FONTSIZE')
  else
  begin
    TB_TrayInfo.FontName := Font;

    if (Size <> '') then
      TB_TrayInfo.FontSize := V;

    TB_TrayInfo.FontStyle := [];

    if (Pos('b', Style) > 0) then
      Include(TB_TrayInfo.FontStyle, fsBold);
    if (Pos('u', Style) > 0) then
      Include(TB_TrayInfo.FontStyle, fsUnderline);
    if (Pos('i', Style) > 0) then
      Include(TB_TrayInfo.FontStyle, fsItalic);
    if (Pos('s', Style) > 0) then
      Include(TB_TrayInfo.FontStyle, fsStrikeout);

    TB_Update;

    Return(data, S_OK);
  end;
end;

// SetTransparent <1|0>
function SetTransparent(mWnd, aWnd: HWND; data, parms: PChar; show, nopause: BOOL): Integer; stdcall;
var
  S: string;
  B: Boolean;
begin
  Result := 3;
  S := string(data);

  // If the value is '1' or 'true', then that
  // means it's transparent, otherwise it's
  // opaque.
  B := (S = '1');

  // If the value has changed from the previous
  // value, we need to update.
  if (B <> TB_TrayInfo.Transparent) then
  begin
    TB_TrayInfo.Transparent := B;
    TB_Update;
  end;

  Return(data, S_OK);
end;

// SetBorder <inset|outset|sunken|raised|etched|none|solid> [rgb]
function SetBorder(mWnd, aWnd: HWND; data, parms: PChar; show, nopause: BOOL): Integer; stdcall;
var
  S: string;
  V, E: Integer;
begin
  Result := 3;
  S := string(data);

  // inset    = BDR_SUNKENOUTER
  // outset   = BDR_RAISEDINNER
  // sunken   = EDGE_SUNKEN
  // raised   = EDGE_RAISED
  // etched   = EDGE_ETCHED
  // NOTE: EDGE_BUMP was not included on purpose.
  // That is one nasty friggin' style and if you
  // want to use it, you're going to have to hax
  // the DLL. =)

  // We default to simple inset.
  TB_TrayInfo.BorderStyle := BDR_SUNKENOUTER;

  if (S = 'outset') then
    TB_TrayInfo.BorderStyle := BDR_RAISEDINNER
  else if (S = 'sunken') then
    TB_TrayInfo.BorderStyle := EDGE_SUNKEN
  else if (S = 'raised') then
    TB_TrayInfo.BorderStyle := EDGE_RAISED
  else if (S = 'etched') then
    TB_TrayInfo.BorderStyle := EDGE_ETCHED
  else if (S = 'none') then
    TB_TrayInfo.BorderStyle := 0
  else if (gettok(S, ' ', 1) = 'solid') then
  begin
    Val(gettok(S, ' ', 2), V, E);

    if (E <> 0) then
    begin
      Return(data, E_INTEGER);
      Exit;
    end;

    TB_TrayInfo.BorderColor := V;
    TB_TrayInfo.BorderStyle := BORDER_SOLID;
  end;

  TB_Update;
  Return(data, S_OK);
end;

// SetWidth <pixels>
function SetWidth(mWnd, aWnd: HWND; data, parms: PChar; show, nopause: BOOL): Integer; stdcall;
var
  V, E: Integer;
begin
  Result := 3;

  // Convert the string into an integer.
  Val(string(data), V, E);

  // Make sure it's a valid integer.
  if (E <> 0) then
    Return(data, E_INTEGER)
  else
  begin
    // Invalidate the area used by the old rectangle, since
    // the new width could be shorter than the old.
    InvalidateRect(TB_Handle, @TB_TrayInfo.UpdateRect, True);

    // Set the new width.
    TB_TrayInfo.Width := V;

    // Get the new update rectangle.
    TB_TrayInfo.UpdateRect := TB_GetUpdateRect;

    // Paint.
    TB_Update;
    Return(data, S_OK);
  end;
end;

// SetColor <rgb integer value>
function SetColor(mWnd, aWnd: HWND; data, parms: PChar; show, nopause: BOOL): Integer; stdcall;
var
  V, E: Integer;
begin
  Result := 3;

  // ...
  Val(string(data), V, E);

  // ...
  if (E <> 0) or (V < 0) or (V > 16777215) then
    Return(data, 'E_INVALID_COLOR')
  else
  begin
    // If it's valid, set the color of the tray text.
    TB_TrayInfo.Color := V;
    TB_Update;
    Return(data, S_OK);
  end;
end;

// SetBkColor <rgb integer value>
function SetBkColor(mWnd, aWnd: HWND; data, parms: PChar; show, nopause: BOOL): Integer; stdcall;
var
  V, E: Integer;
begin
  Result := 3;

  // ...
  Val(string(data), V, E);

  // ...
  if (E <> 0) or (V < 0) or (V > 16777215) then
    Return(data, 'E_INVALID_COLOR')
  else
  begin
    // If it's valid, set the color of the tray background.
    TB_TrayInfo.BkColor := V;
    TB_Update;
    Return(data, S_OK);
  end;
end;

// SetText <text string>
function SetText(mWnd, aWnd: HWND; data, parms: PChar; show, nopause: BOOL): Integer; stdcall;
begin
  Result := 3;

  // Set the new text and update.
  TB_TrayInfo.Text := string(data);
  TB_Update;

  Return(data, S_OK);
end;

// SetMargins <top> <right> <bottom>
function SetMargins(mWnd, aWnd: HWND; data, parms: PChar; show, nopause: BOOL): Integer; stdcall;
var
  S: string;
  V1, V2, V3: Integer;
  E1, E2, E3: Integer;
begin
  Result := 3;
  S := string(data);

  // Go ahead and assume they entered an invalid integer. :)
  Return(data, E_INTEGER);

  // Check all three values...
  Val(gettok(S, ' ', 1), V1, E1);
  Val(gettok(S, ' ', 2), V2, E2);
  Val(gettok(S, ' ', 3), V3, E3);

  // If any screwed up, get out now.
  if (E1 <> 0) or (E2 <> 0) or (E3 <> 0) then
    Exit;

  // Since changing the margin is going to require a repaint, we invalidate
  // the area used by the rectangle.
  InvalidateRect(TB_Handle, @TB_TrayInfo.UpdateRect, True);

  // We save the new margins...
  TB_TrayInfo.MarginTop := V1;
  TB_TrayInfo.MarginRight := V2;
  TB_TrayInfo.MarginBottom := V3;

  // We get the updated paint rectangle...
  TB_TrayInfo.UpdateRect := TB_GetUpdateRect;

  // And we paint.
  TB_Update;

  Return(data, S_OK);
end;

// Tooltip [-dc] [delay] [text]
function Tooltip(mWnd, aWnd: HWND; data, parms: PChar; show, nopause: BOOL): Integer; stdcall;
var
  S: string;
  SW: string;
  V, E: Integer;
begin
  Result := 3;
  S := string(data);

  // Get the first word, most likely a switch.
  SW := gettok(S, ' ', 1);

  // If it's not a switch, then it's text, and we just go ahead
  // and set the text here.
  if (SW[1] <> '-') then
    TT_Text := S

  // If it's -c, we clear the text.
  else if (SW = '-c') then
    TT_Text := ''

  // If it's -d, it's delay, but there's also an optional [text] that can
  // be there.
  else if (SW = '-d') then
  begin
    // Convert the delay into an integer.  We can't use intstr() here because
    // we don't know what this is.  It could be "adkf" and that's definately
    // not an integer.
    Val(gettok(S, ' ', 2), V, E);

    // If it's no good, just return an error and get out.
    if (E <> 0) then
    begin
      Return(data, 'E_INVALID_INTEGER');
      Exit;
    end;

    // If it's good, save it.
    TT_Delay := V;

    // If there is any text there, we need to set the new tooltip, but only if
    // there is text there.  When specifying a negative number with gettok(),
    // it grabs everything from N to the end (like N- in $gettok).
    if (gettok(S, ' ', 3) <> '') then
      TT_Text := gettok(S, ' ', -3);
  end;

  Return(data, S_OK);
end;

function TB_SubClass(hWnd: HWND; Msg: Cardinal; wParam: Integer; lParam: Integer): Integer; stdcall;
var
  P: TPoint;
  R: TRect;
begin
  // Get rid of that friggin "result of TB_SubClass may be undefined"
  // warning.  We've covered all the possible cases...
  Result := 0;

  // If we're exiting, we need to clean up.
  if (Msg = WM_DESTROY) then
    TB_Uninitialize;

  // tbwin.dll has just been loaded or unloaded.  That means the window
  // procedure the toolbar has changed.  We need to update our saved one, so
  // if we unload stuff don't mess up.
  if (Msg = WM_NOTIFYTBWIN) then
  begin
    if not (LoadedFirst) and (wParam <> 0) then
    begin
      TB_OldProc := Pointer(wParam);
      Result := Integer(@TB_SubClass);
    end
    else
      Result := Integer(TB_WndProc);

    if (lParam = 1) then
      LoadedFirst := True;
    Exit;
  end;

  // Find the x/y coordinates of the mouse to check for the mouse
  // events, if we need to.
  P.x := LOWORD(lParam);
  P.y := HIWORD(lParam);

  // Basically, we only want to stop the default processing if there
  // is a right-click (show menu), double-click (dclick item), or
  // single-click (sclick item).
  if ((Msg = WM_RBUTTONDOWN) or (Msg = WM_LBUTTONDBLCLK) or (Msg = WM_LBUTTONUP)) and (PtInRect(TB_TrayInfo.UpdateRect, P)) then
  begin
    // We need to hide the tooltip (to get it out of the way).  Also, after
    // hiding it, a WM_MOUSEMOVE is called, so we need to set a flag so that
    // the tooltip doesn't pop back up right after we click.
    // Don't let it accidentally pop back up!
    TT_Hide;
    NoMouseMove := True;

    case Msg of
      // right-click
      WM_RBUTTONDOWN:
        // If there's any menu items, show the menu.
        // Make sure to account for the offset of
        // the toolbar, since the points are in screen
        // coordinates and we need client coordinates.
        if (TB_MenuItemCount > 0) then
        begin
          GetWindowRect(TB_Handle, R);
          TrackPopupMenuEx(TB_Menu, 0, R.Left + P.x, R.Top + P.y, TB_Handle, nil);
        end;

      // double-click
      WM_LBUTTONDBLCLK:
        // If there's a command, execute it.
        if (TB_MenuDclick <> '') then
          ExecCommand(TB_MenuDclick);

      // single-click
      WM_LBUTTONUP:
        // If there's a command, execute it.
        if (TB_MenuSclick <> '') then
          ExecCommand(TB_MenuSclick);
    end;
  end
  else
  // Otherwise, we want to call the original window procedure and
  // let mIRC do his thang.
    Result := CallWindowProc(TB_OldProc, hWnd, Msg, wParam, lParam);

  // We use this to trap when a user clicks a menu item in our
  // custom menu.  We then find out what item it was, match it up
  // with our array, and execute the appropriate command.
  if (Msg = WM_COMMAND) and (HIWORD(wParam) = 0) and ((LOWORD(wParam) - TB_MenuStart) >= 1) then
    ExecCommand(TB_MenuItems[LOWORD(wParam) - TB_MenuStart])

  // If the user is sizing mIRC, we need to make sure the tray gets
  // updated as well.
  else if (Msg = WM_SIZE) then
  begin
    InvalidateRect(TB_Handle, @TB_TrayInfo.UpdateRect, True);
    TB_TrayInfo.UpdateRect := TB_GetUpdateRect;
  end

  else if (Msg = WM_MOUSEMOVE) and (PtInRect(TB_TrayInfo.UpdateRect, P)) then
  begin
    if not (IsWindow(TT_Handle)) and (GetForegroundWindow = MainHandle) and (TT_Text <> '') and not (NoMouseMove) then
    begin
      KillTimer(TB_Handle, TT_TIMER);
      SetTimer(TB_Handle, TT_TIMER, TT_Delay, nil);
    end;
    NoMouseMove := False;
  end

  else if (Msg = WM_TIMER) and (wParam = TT_TIMER) then
  begin
    KillTimer(TB_Handle, TT_TIMER);
    GetCursorPos(P);
    ScreenToClient(TB_Handle, P);

    if not (IsWindow(TT_Handle)) and (TT_Text <> '') and (PtInRect(TB_GetUpdateRect, P)) then
    begin
      TT_Show;
      TT_Update(P);
      SetTimer(TT_Handle, 1, 1, nil);
    end;
  end

  // Hey, paint the tray too! :)
  else if (Msg = WM_PAINT) then
    TB_UpdatePaint;
end;

procedure TB_ClearMenu;
var
  I: Integer;
begin
  // Free memory used for menu items.
  for I := 1 to TB_MenuItemCount do
    DeleteMenu(TB_Menu, TB_MenuStart + I, MF_BYCOMMAND);

  // Reset dclick menu.
  TB_MenuDclick := '';

  // Reset sclick menu.
  TB_MenuSclick := '';

  // Reset number of menu items.
  TB_MenuItemCount := 0;
end;

{ This function updates without a corresponding WM_PAINT message. }
procedure TB_Update;
begin
  // Invalidate the rectangle used by the tray.  If it's transparent,
  // we need to also erase the area used.
  InvalidateRect(TB_Handle, @TB_TrayInfo.UpdateRect, TB_TrayInfo.Transparent);

  // Update the tray area.
  UpdateWindow(TB_Handle);
end;

{ This function MUST be called from inside the WM_PAINT message. }
procedure TB_UpdatePaint;
const
  BoldText: array[Boolean] of Integer = (FW_NORMAL, FW_BOLD);
var
  DC: HDC;
  Font: HFONT;
  Pen: HPEN;
  Brush: HBRUSH;
  W, H: Integer;
  R: TRect;
  SZ: TSize;
  DrawMode: DWORD;
begin
  if (TB_Handle = 0) or not (IsWindowVisible(TB_Handle)) then
    Exit;

  // Get the device to paint on.
  DC := GetDC(TB_Handle);

  // Get a font to paint with. :)
  Font := CreateFont(MulDiv(TB_TrayInfo.FontSize, GetDeviceCaps(DC, LOGPIXELSY), 72), 0, 0, 0, BoldText[fsBold in TB_TrayInfo.FontStyle], Ord(fsItalic in TB_TrayInfo.FontStyle), Ord(fsUnderline in TB_TrayInfo.FontStyle), Ord(fsStrikeout in TB_TrayInfo.FontStyle), DEFAULT_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY, DEFAULT_PITCH or FF_DONTCARE, PChar(TB_TrayInfo.FontName));

  // Shortcut so I don't have to do TB_TrayInfo.UpdateRect.*
  R := TB_TrayInfo.UpdateRect;

  // Save the old font and tell the device to use the new font.
  Font := SelectObject(DC, Font);

  // Set the appropriate colors.
  Windows.SetBkColor(DC, TB_TrayInfo.BkColor);
  SetTextColor(DC, TB_TrayInfo.Color);

  // Get the width and height of the text so we can center it.
  GetTextExtentPoint32(DC, PChar(TB_TrayInfo.Text), Length(TB_TrayInfo.Text), SZ);

  // Set the mode depending on if we have transparent set or not.
  // You'll probably want to always use opaque, not transparent,
  // because the text is harder to read with a colorful background
  // bitmap and it's useless if you have no background bitmap.
  if (TB_TrayInfo.Transparent) then
  begin
    SetBkMode(DC, TRANSPARENT);
    DrawMode := 0
  end
  else
  begin
    SetBkMode(DC, OPAQUE);
    DrawMode := ETO_OPAQUE;
  end;

  // shortcuts
  W := R.Right - R.Left;
  H := R.Bottom - R.Top;

  // Draw the text -- ExtTextOut is nice because it can center and clip
  // and all that good stuff in one go.  DrawTextEx can be used for
  // similar results.
  ExtTextOut(DC, R.Left + (W - SZ.cx) div 2, R.Top + (H - SZ.cy) div 2, ETO_CLIPPED or DrawMode, @R, PChar(TB_TrayInfo.Text), Length(TB_TrayInfo.Text), nil);

  // Draw the solid border.
  if (TB_TrayInfo.BorderStyle = BORDER_SOLID) then
  begin
    // Create the solid pen to draw the border.
    Pen := CreatePen(PS_SOLID, 1, TB_TrayInfo.BorderColor);

    // Save the old pen, very important.
    Pen := SelectObject(DC, Pen);

    // Make the brush hollow or clear.  We only want to draw
    // a border, not fill the rectangle.  The text output has
    // already been done, so filling here wouldn't be good. Also
    // save the old brush so we can restore it.
    Brush := SelectObject(DC, GetStockObject(HOLLOW_BRUSH));

    // Draw the border with the solid pen and clear brush.
    Rectangle(DC, R.Left, R.Top, R.Right, R.Bottom);

    // Restore the old pen and get the handle of the one we created.
    Pen := SelectObject(DC, Pen);

    // Free the memory used by the pen *we* created.
    DeleteObject(Pen);

    // Restore the old brush.  Since we used a stock object (HOLLOW_BRUSH),
    // we don't have to free the memory.  The stock brushes are all handled
    // by Windows.
    SelectObject(DC, Brush);
  end
  // Otherwise, draw the beveled edge.
  else if (TB_TrayInfo.BorderStyle > 0) then
    // DrawEdge does basically everything for us.
    DrawEdge(DC, R, TB_TrayInfo.BorderStyle, BF_RECT);

  // Select the original font.
  Font := SelectObject(DC, Font);

  // Free the one we created.
  DeleteObject(Font);

  // Release the device from painting.
  ReleaseDC(TB_Handle, DC);
end;

function TB_GetUpdateRect: TRect;
begin
  // This gets the update rectangle of the current tray.
  GetClientRect(TB_Handle, Result);

  // Take into account the width...
  Result.Left := Result.Right - TB_TrayInfo.Width;

  // ...and the margins.
  OffsetRect(Result, -TB_TrayInfo.MarginRight, 0);
  Inc(Result.Top, TB_TrayInfo.MarginTop);
  Dec(Result.Bottom, TB_TrayInfo.MarginBottom);
end;

function TB_Initialize(mWnd: HWND): string;
var
  H: HWND;
  Res: Integer;
begin
  // Let's find the handle of the toolbar.  This is
  // pretty simple because the class name is very unique.
  H := FindWindowEx(mWnd, 0, 'mIRC_Toolbar', '');

  if not (IsWindow(H)) then
    H := FindWindowEx(mWnd, 0, 'mIRCTB32', '');

  if (H = TB_Handle) and (IsWindow(H)) then
  begin
    // We are already initialized.. just exit now.
    Result := S_OK;
    Exit;
  end;

  TB_Handle := H;

  if (TB_Handle = 0) then
    Result := 'E_TOOLBAR_NOT_FOUND'
  else
  begin
    // tbwin.dll/tbtray.dll compatibility code.
    Res := SendMessage(TB_Handle, WM_NOTIFYTBTRAY, 0, 0);
    TB_OldProc := NewMessageProc(TB_Handle, @TB_SubClass);
    LoadedFirst := (Res = 0);
    if (LoadedFirst) then
      TB_WndProc := TB_OldProc
    else
      TB_WndProc := Pointer(Res);

    // Save the old class style.
    TB_ClassStyle := GetClassLong(TB_Handle, GCL_STYLE);

    // We have to add the CS_DBLCLKS style to use the "dclick" menu item
    // because toolbars don't have it by default.
    SetClassLong(TB_Handle, GCL_STYLE, TB_ClassStyle or CS_DBLCLKS);

    // Set the current rectangle and update when we load.
    TB_TrayInfo.UpdateRect := TB_GetUpdateRect;
    TB_Update;
    Result := S_OK;
  end;
end;

function TB_Uninitialize: string;
var
  R: TRect;
  Res: Integer;
begin
  if (TB_OldProc <> nil) and (IsWindow(TB_Handle)) then
  begin

    // tbwin.dll/tbtray.dll compat. code.
    Res := 0;
    if (LoadedFirst) then
      Res := SendMessage(TB_Handle, WM_NOTIFYTBTRAY, Integer(TB_WndProc), 1);

    if (Res <> 0) then
      RestoreMessageProc(TB_Handle, Pointer(Res))
    else
      RestoreMessageProc(TB_Handle, TB_OldProc);

    // Restore the old class style, probably doesn't matter but we
    // do it anyways, in case some other DLL modified it too.
    SetClassLong(TB_Handle, GCL_STYLE, TB_ClassStyle);

    // Clear the area used by the tray.
    GetClientRect(TB_Handle, R);
    InvalidateRect(TB_Handle, @R, True);
    UpdateWindow(TB_Handle);
  end;

  // Free the individual memory used by the menu items.
  TB_ClearMenu;

  // Free the menu itself.
  DestroyMenu(TB_Menu);

  // Cleanup.
  TB_Menu := 0;
  TB_Handle := 0;
  TB_OldProc := nil;

  // Remove tooltip if it's there.
  TT_Hide;

  Result := S_OK;
end;

exports
  LoadDll, UnloadDll,
  OpenTray, CloseTray,
  AddMenu, ClearMenu,
  SetFont, SetText, SetColor,
  SetBkColor,
  SetMargins,
  SetBorder,
  SetTransparent,
  SetWidth,
  Tooltip;

begin
  MainHandle := 0;
  NeedUnload := False;
end.
 